#include <JTC\JTC.h>
#include <iostream.h>
#include <fstream.h>

class DatoCondiviso : public JTCMonitor

{
	int dato;
	int numLettori;
	bool ceUnoScrittore;

public:
	DatoCondiviso()
	{
		dato = 0;
		numLettori = 0;
		ceUnoScrittore = false;
	}
	void cominciaLettura()
	{
		JTCSynchronized sync(*this);
		while(ceUnoScrittore)
			wait();
		++numLettori;
	};
	void terminaLettura()
	{
		JTCSynchronized sync(*this);
		--numLettori;
		if (numLettori == 0)
			notify();
	};
	void cominciaScrittura ()
	{
		JTCSynchronized sync(*this);
		while (numLettori > 0 || ceUnoScrittore)
			wait();
		ceUnoScrittore = true;
	};
	void terminaScrittura()
	{
		JTCSynchronized sync(*this);
		ceUnoScrittore = false;
		notifyAll();
	};
	int leggiDato()
	{
		return dato;
	}
	void scriviDato(int i)
	{
		dato = i;
	}
};

// Versione senza starvation per gli scrittori
#define SOGLIAGIRI 5
// SOGLIAGIRI = Numero di lettori che possono al massimo passare davanti a uno scrittore


class DatoCondivisoSenzaStarvation : public DatoCondiviso

{
	int dato;
	int numLettori;
	int numScrittoriInAttesa;
	int numGiriSenzaScrittori;
	bool ceUnoScrittore;

public:
	DatoCondivisoSenzaStarvation()
	{
		dato = 0;
		numLettori = 0;
		ceUnoScrittore = false;
		numScrittoriInAttesa = 0;
		numGiriSenzaScrittori;
	}
	void cominciaLettura()
	{
		JTCSynchronized sync(*this);
		while(ceUnoScrittore || (numScrittoriInAttesa > 0 && numGiriSenzaScrittori > SOGLIAGIRI))
			wait();
		++numLettori;
		++numGiriSenzaScrittori;
	};
	void terminaLettura()
	{
		JTCSynchronized sync(*this);
		--numLettori;
		if (numLettori == 0)
			notify();
	};
	void cominciaScrittura ()
	{
		JTCSynchronized sync(*this);
		numScrittoriInAttesa++;
		while (numLettori > 0 || ceUnoScrittore)
			wait();
		ceUnoScrittore = true;
		numScrittoriInAttesa--;
		numGiriSenzaScrittori = 0;
	};
	void terminaScrittura()
	{
		JTCSynchronized sync(*this);
		ceUnoScrittore = false;
		notifyAll();
	};
	int leggiDato()
	{
		return dato;
	}
	void scriviDato(int i)
	{
		dato = i;
	}
};

DatoCondivisoSenzaStarvation f;


class Scrittore : public JTCThread

{
	int id;
public:
	Scrittore(int i)
	{
		id = i;
	}
	virtual void run()
	{
		while(true) {
			cout << "Lo scrittore " << id << " Chiede di scrivere " << endl;
			f.cominciaScrittura();		
			cout << "Lo scrittore " << id << " Comincia a scrivere " << endl;
			sleep(rand() % 1000);
			f.scriviDato(id);
			cout << "Lo scrittore " << id << " Ha scritto " << id << endl;
			f.terminaScrittura();
			cout << "Lo scrittore " << id << " termina di scrivere " << endl;
			sleep(rand() % 5000);
		}
	}
};

class Lettore : public JTCThread

{
	int id;
public:
	Lettore(int i)
	{
		id = i;
	}
	virtual void run()
	{
		while(true) {
			cout << "Il lettore " << id << " Chiede di leggere " << endl;
			f.cominciaLettura();		
			cout << "Il lettore " << id << " Comincia a leggere " << endl;
			sleep(rand() % 100);
			cout << "Il lettore " << id << " legge " << f.leggiDato() << endl;
			f.terminaLettura();
			cout << "Il lettore " << id << " termina di leggere " << endl;
			sleep(rand() % 5000);
		}
	}
};

int main(int argc, char* argv[])

{   
	JTCInitialize initialize;
	Scrittore* p[2];
	Lettore* c[4];
    for (int i = 0; i < 2; i++) {
		p[i] = new Scrittore(i);
		p[i]->start();
	}
	for (i = 0; i < 4; i++) {
		c[i] = new Lettore(i);
		c[i]->start();
	}
	return 0;	
}

